/*
*****************************************************************************
*
*      File          : tfo_decision.c We consider version handling no longer;
*      NOTE          : it is not expected that this SW will ever change,
*                      therefore all version handling is removed.
*      Version       : C.0.0, 02. December 2014, Karl Hellwig, Ericsson
*      Version       : B.0.0, 18. September 2012, Paolo Usai, Karl Hellwig
*      Version       : A.0.0, 13. April 2011, Karl Hellwig
*      Version       : 9.0.0, 11. December 2009, Paolo Usai
*      Version       : 8.1.0, 27. March 2009, Karl Hellwig
*      Version       : 7.0.0, 14. June 2007, Paolo Usai
*      Version       : 6.4.0, 05. June 2007, Karl Hellwig
*      Version       : 6.3.0, 02. October 2006, Karl Hellwig
*      Version       : 6.2.0, 30. August 2005, Karl Hellwig
*      Version       : 2.2.0, 27. November 2000, Jochen Junkawitsch
*      Description   : source file for TFO Decision algorithm of 3GPP TS 28.062
*
*****************************************************************************
*/

#include <stdio.h>
#include <stdlib.h>
#include "tfo_decision.h"	/* KHe: consider version here no longer*/
/*#include <iostream.h> */


/* ---------------------------------------------------------------------------------
                     Main TFO Decision Procedure (section 11.1)
   --------------------------------------------------------------------------------- */

int tfo_decision_main(Data* data)
{

/* ---------------------------------------------------------------------------------
   Case: Two AMR-WB Codec Types with FR_AMR-WB or OHR_AMR-WB on either side

   Here only the three lower modes are possible and they are always available.
   
   oACS = iACS = cACS = (12.65 + 8.85 + 6.60)
                      =   4    +  2   +  1    = 7
   
   No Modification is possible or necessary. Immediate TFO always.
   --------------------------------------------------------------------------------- */

    if  (data->lType ==   FR_AMR_WB && data->dType ==   FR_AMR_WB ||
		 data->lType ==  OHR_AMR_WB && data->dType ==  OHR_AMR_WB ||
		
		 data->lType ==   FR_AMR_WB && data->dType == UMTS_AMR_WB ||
		 data->lType ==   FR_AMR_WB && data->dType ==  OFR_AMR_WB ||
		 data->lType ==   FR_AMR_WB && data->dType ==  OHR_AMR_WB ||

         data->lType ==  OHR_AMR_WB && data->dType == UMTS_AMR_WB ||
		 data->lType ==  OHR_AMR_WB && data->dType ==  OFR_AMR_WB ||
		 data->lType ==  OHR_AMR_WB && data->dType ==   FR_AMR_WB ||

         data->lType == UMTS_AMR_WB && data->dType ==   FR_AMR_WB ||
		 data->lType ==  OFR_AMR_WB && data->dType ==   FR_AMR_WB ||
		 data->lType ==  OHR_AMR_WB && data->dType ==   FR_AMR_WB ||

         data->lType == UMTS_AMR_WB && data->dType ==  OHR_AMR_WB ||
		 data->lType ==  OFR_AMR_WB && data->dType ==  OHR_AMR_WB ||
		 data->lType ==   FR_AMR_WB && data->dType ==  OHR_AMR_WB   )
  {
		data->cacs = data->cscs = data->oacs = data->iacs = 7;
	    printf("Immediate TFO - no Modification necessary\n");
		return 1;
  }

/* ---------------------------------------------------------------------------------
   Case: Two AMR-WB Codec Types with neither FR_AMR-WB nor OHR_AM-WB.

		All 6*6 combinations of Allowed Configurations can occur!

		The  cACS = lACS & dACS   are easy to calculate
		The  cSCS = lSCS & dSCS   also, but this is not important.

		The good thing: immediate TFO is always possible for AMR-WB! 
		
		iACS = cACS =  [23.85] / [15.85] + (12.65 + 8.85 + 6.60)
		            =   [16]   /   [8]   +    4   +  2  +   1    = 23 / 15 / 7
		
		The oACS must be out of these three values, too.
															
		The only questions: what is the oACS? And which side must adapt?				
  --------------------------------------------------------------------------------- */

  if (data->lType == UMTS_AMR_WB && data->dType == UMTS_AMR_WB ||
	  data->lType == UMTS_AMR_WB && data->dType ==  OFR_AMR_WB ||
	  data->lType ==  OFR_AMR_WB && data->dType == UMTS_AMR_WB ||
	  data->lType ==  OFR_AMR_WB && data->dType ==  OFR_AMR_WB   )
  {
	  printf("Immediate TFO ");
	  data->cacs = (data->lacs & data->dacs);		// thats always true
	  data->cscs = (data->lscs & data->dscs);		// thats always true
	  data->iacs =  data->cacs;						// thats always true
	  
	  if (data->cacs > 7)							// both sides support the same upper configuration
	  {
		  data->oacs = data->cacs;					// and we are ready
		  printf("- no Modification necessary\n");	// regardless whether modifications are supported or not
		  return 1;
	  }

	  if ((data->lOM == 0) && (data->dOM == 0))		// no side does support modification
	  { 
		  data->oacs = 7;							// and we are ready with 3 lower modes
		  printf("- no Modification necessary\n");
		  return 1;
	  }

	  if ((data->lOM == 1) && (data->dOM == 0))		// only the local side supports modification
	  { 
		  data->oacs = data->dacs;					// and we are ready with dACS

		  printf("- then Modification of local Configuration\n");
		  return 1;
	  }

	  if ((data->lOM == 0) && (data->dOM == 1))		// only the distant side supports modification
	  { 
		  data->oacs = data->lacs;					// and we are ready with lACS

		  printf("- then Modification of distant Configuration\n");
		  return 1;
	  }

	  if (data->cscs == 31)							// both sides support modification, but are different
	  {
		if (data->lacs > data->dacs)				// which one wins? the ACS with the highest mode!
		{
			data->oacs = data->lacs;
			printf("- then Modification of distant Configuration\n");
		}
		if (data->lacs < data->dacs)
		{
			data->oacs = data->dacs;
			printf("- then Modification of local Configuration\n");
		}		  

	  }

	  return 1;	// O.K. AMR-WB is done
  }

/* ---------------------------------------------------------------------------------
   Case: Two compatible AMR Codec Types
   --------------------------------------------------------------------------------- */

  if (data->lType == UMTS_AMR   && data->dType == UMTS_AMR   ||	// the exotic beast first

	  data->lType == UMTS_AMR_2 && data->dType == UMTS_AMR_2 ||
      data->lType == UMTS_AMR_2 && data->dType == FR_AMR     ||
      data->lType == UMTS_AMR_2 && data->dType == OHR_AMR    ||
      data->lType == UMTS_AMR_2 && data->dType == HR_AMR     ||
	  
      data->lType == FR_AMR     && data->dType == UMTS_AMR_2 ||
      data->lType == FR_AMR     && data->dType == FR_AMR     ||
      data->lType == FR_AMR     && data->dType == OHR_AMR    ||
      data->lType == FR_AMR     && data->dType == HR_AMR     ||

      data->lType == OHR_AMR    && data->dType == UMTS_AMR_2 ||
      data->lType == OHR_AMR    && data->dType == FR_AMR     ||
      data->lType == OHR_AMR    && data->dType == OHR_AMR    ||
      data->lType == OHR_AMR    && data->dType == HR_AMR     ||

      data->lType == HR_AMR     && data->dType == UMTS_AMR_2 ||
      data->lType == HR_AMR     && data->dType == FR_AMR     ||
	  data->lType == HR_AMR     && data->dType == OHR_AMR    ||
      data->lType == HR_AMR     && data->dType == HR_AMR
	 )
  {
	determine_IACS   (data);	// first determine iACS and oACS
	determine_OACS   (data);	// only necessary for the AMR family

	if (data->lacs == data->dacs)						// the ideal case: both sides are identical 
	{
		printf("Immediate TFO on iACS; no further optimisation\n\n"); 
		return 0;
	}


	if (FR_HR_matching(data))							// maybe not identical, but well compatible
	{													// mainly for FR and HR constellations
		printf("Immediate TFO on iACS; FR-HR Matching; no further optimisation\n\n");
		return 0;
	}


	if (!acceptable(data, data->oacs))				// o..o..o! The oACS is not acceptable
	{
		if ( (data->lOM == 1) || (data->dOM == 1) ) 
			printf("First Codec Mismatch Resolution, then TFO might be possible.\n");

		else 
			printf("TFO not possible for various reasons.\n");
		return 0;
	}

		
	if (data->cacs == data->oacs)						// the oACS is identical to the common ACS
	{
		printf("Immediate TFO on iACS\n\n"); return 0;
	}

	if (contiguous_subset(data->cacs, data->oacs))	// the cACS is not perfect, but allows TFO setup
	{													// with subsequent handover of one or both sides to oACS
		printf("Immediate TFO on iACS, then Codec Mode Optimization to oACS.\n\n");
		return 0;
	}

	printf("TFO is possible after change to oACS.\n");// no immediate TFO, first change to oACS
	return 0;						

  }

/* ---------------------------------------------------------------------------------
   Case: UMTS_AMR <-> UMTS_AMR_2	: that's new after 4.5.0

   Since December 2004 the UMTS_AMR is considered as incompatible to all Codec Types,
   also to UMTS_AMR_2, when the Configuration contains more than one mode.
   Reason: Rate Control timing.
   For Single-Mode Configurations UMTS_AMR would be compatile too ALL AMR family members,
   But it was agreed to allow only UMTS_AMR_2 compatibility even for Single Mode.
   
   Basically the solution was constructed here again. 
   This was simpler, than to take existing SW blocks.
   The decison follows that of the other Single Mode cases (cMACS == 1).

   Only the SCSes, but not the ACSses are considered (as in cMACS == 1 in general). 
   This is not the most reasonable logic. But it is standard.
   It leads too often to strange results, with too low a Codec Mode.

   No further effort is considered necessary, 
   because this situation should anyway never be used: 
   Opertors Guide: Use UMTS_AMR only in Single Mode 12.2.

   --------------------------------------------------------------------------------- */

  if (data->lType == UMTS_AMR_2 && data->dType == UMTS_AMR   ||
      data->lType == UMTS_AMR   && data->dType == UMTS_AMR_2   )
  {
	  determine_IACS   (data);	// first determine iACS and oACS
	  determine_OACS   (data);	// only necessary for the AMR family

	  printf("Since Dec 2004: UMTS_AMR and UMTS_AMR_2 are only compatible \nin single mode configurations! \n\n");

	  data->cacs = (data->lacs) & (data->dacs);

	  if ((nModes (data->cacs) == 1) && (data->lacs == data->dacs))
	  { 
		  printf("Immediate TFO: Exactly one mode, common to both sides\n\n");
	      return 0;
	  }

	  if ((data->lOM == 0) && (data->dOM == 0))
	  {
		  printf("TFO is not possible: No commmon valid Single-Mode Configuration, no side can adapt\n\n");
		  return 0;
	  }

	  if ((data->lOM == 1) && (data->dOM == 0))
	  {
		  if (nModes (data->dacs) != 1)
		  {
			 printf("TFO is not possible: Distant side has more than one mode and can not adapt\n");
			 printf("The case that Rate Control could exclude the higher mode is not considered!\n\n");

			 return 0;
		  }
	      data->cacs = data->lscs & data->dacs;
		  if (nModes (data->cacs) == 1)
		  {
			 printf("TFO is possible after Modification: Local side can adapt to Distant side\n\n");
			 return 0;
		  }
		  printf("No common mode\nTFO is not possible\n\n");
	  }

	  if ((data->lOM == 0) && (data->dOM == 1))
	  {
		  if (nModes (data->lacs) != 1)
		  {
			 printf("TFO is not possible: Local side has more than one mode and can not adapt\n\n");
			 printf("The case that Rate Control could exclude the higher mode is not considered!\n\n");

			 return 0;
		  }
	      data->cacs = data->lacs & data->dscs;
		  if (nModes (data->cacs) == 1)
		  {
			 printf("TFO is possible after Modification: Distant side can adapt to Local side\n\n");
			 return 0;
		  }
		  printf("TFO is not possible: No common mode\n\n");
	  }

	  if ((data->lOM == 1) && (data->dOM == 1))
	  {
		  printf("Both sides can adapt, this is the most difficult decision\n");
	      printf("Select one mode according to preference list: \n 6.7  7.4  5.9  5.15  4.75  7.95  10.2  12.2\n");
	      data->cscs = (data->lscs) & (data->dscs);
		  if (data->oacs = data->cscs & mode_670) {printf("TFO is possible after Modification\n\n"); return 0;}
		  if (data->oacs = data->cscs & mode_740) {printf("TFO is possible after Modification\n\n"); return 0;}
		  if (data->oacs = data->cscs & mode_590) {printf("TFO is possible after Modification\n\n"); return 0;}
		  if (data->oacs = data->cscs & mode_515) {printf("TFO is possible after Modification\n\n"); return 0;}
		  if (data->oacs = data->cscs & mode_475) {printf("TFO is possible after Modification\n\n"); return 0;}
		  if (data->oacs = data->cscs & mode_795) {printf("TFO is possible after Modification\n\n"); return 0;}
		  if (data->oacs = data->cscs & mode_102) {printf("TFO is possible after Modification\n\n"); return 0;}
		  if (data->oacs = data->cscs & mode_122) {printf("TFO is possible after Modification\n\n"); return 0;}
	  }

	  printf("TFO is not possible\n\n"); return 0; // done with AMR family
  }

/* ---------------------------------------------------------------------------------
   Case: Two compatible simple codec types
   --------------------------------------------------------------------------------- */
 
  if (data->lType == GSM_FR  && data->dType == GSM_FR  ||
      data->lType == GSM_EFR && data->dType == GSM_EFR ||
      data->lType == GSM_HR  && data->dType == GSM_HR    ) 
  {
	  printf("Immediate TFO\n\n");
	  return 0;
  }

/* ---------------------------------------------------------------------------------
   Case: The rest is Codec Mismatch	
   --------------------------------------------------------------------------------- */
  printf("TFO is not possible\n\n");
  return 0;

}



/* ---------------------------------------------------------------------------------
   FR - HR - Matching (section 12.6)
   --------------------------------------------------------------------------------- */

int FR_HR_matching(Data* data)
{
  if (!(data->lType == FR_AMR     && data->dType == HR_AMR     ||
		data->lType == OHR_AMR    && data->dType == HR_AMR     ||
		data->lType == UMTS_AMR_2 && data->dType == HR_AMR     ||
		data->lType == HR_AMR     && data->dType == FR_AMR     ||
		data->lType == HR_AMR     && data->dType == OHR_AMR    ||
		data->lType == HR_AMR     && data->dType == UMTS_AMR_2   )) 
      return 0;				// No FR-HR and no HR-FR match

  if (data->cacs == 0)
	  return 0;				// No common ACS						

  if (!acceptable(data, data->cacs))
	  return 0;	

  return 1;
}



/* ---------------------------------------------------------------------------------
   Determine_IACS                     
   --------------------------------------------------------------------------------- */

int determine_IACS(Data* data)
{

  data->cscs  = data->lscs & data->dscs;

  data->cmacs = data->lmacs < data->dmacs ? data->lmacs : data->dmacs;	// minimum

  data->cacs  = contiguous_subset_selection ((CodecSet)(data->lacs & data->dacs), 0,0,
					                          data->lacs, data->dacs, data->cmacs);
  return 0;
}



/* ---------------------------------------------------------------------------------
   Determine_OACS                     
   --------------------------------------------------------------------------------- */

int determine_OACS (Data* data)
{
/* ---------------------------------------------------------------------------------
   Case: no side supports ACS change                     
   --------------------------------------------------------------------------------- */
  if (!data->lOM && !data->dOM)
    calc_OACS_case1 (data);

/* ---------------------------------------------------------------------------------
   Case: both sides support ACS change                     
   --------------------------------------------------------------------------------- */
  else if (data->lOM && data->dOM)
    /* calculate OACS */
    data->oacs = calc_OACS (data->cscs, data->cmacs,
			                data->lType == HR_AMR || data->dType == HR_AMR ? HR_AMR : FR_AMR);

/* ---------------------------------------------------------------------------------
   Case: only one side supports ACS change                     
   --------------------------------------------------------------------------------- */

  else
    calc_OACS_case2 (data);

  return 0;
}



/* ---------------------------------------------------------------------------------
   Is the oACS acceptable, when compared to the original ACSes?                     
   --------------------------------------------------------------------------------- */

int acceptable(Data* data, CodecSet cs)
{
	// first check if either of both sides uses a HR_AMR, because HR_AMR as stronger restrictions
	// if no HR_AMR is involved, then FR_AMR rules apply, also for UMTS_AMR, UMTS_AMR2 and OHR_AMR.

	CodecType codecType = data->lType == HR_AMR || data->dType == HR_AMR ? HR_AMR : FR_AMR;

  return lowModeRule  (data->lacs, data->dacs, cs, codecType) &&
         highModeRule (data->lacs, data->dacs, cs);;
}



/* ---------------------------------------------------------------------------------
	check if "a" is a contiguous subset of "b"
   --------------------------------------------------------------------------------- */

int contiguous_subset(CodecSet a, CodecSet b)
{
  int mode;

  if (a == 0) return 0;					// subset is empty

  for (mode=mode_122; mode; mode>>=1)
  {
    if (a & mode) break;				// look for the highest mode in "a" first
										// delete all bits in "b" which are higher than in "a"
    b &= (mode ^ 0xFF);					// read that as:  b = b & (mode bitwise inverted)
  }

  return a == b;						// if now a==b then a is fully included in b
}										// i.e. "a" is a contiguous subset of "b" or "a" is identical to "b"


/* ---------------------------------------------------------------------------------
	section 12.2.2 Case 1: No side supports ACS change
   --------------------------------------------------------------------------------- */

int calc_OACS_case1 (Data* data)
{
  CodecSet cacs = data->lacs & data->dacs;

  if (!contiguous_subset(cacs, data->lacs) || !contiguous_subset(cacs, data->dacs))
  {
    data->oacs = contiguous_subset_selection (cacs, data->lOM, data->dOM,
											  data->lacs, data->dacs, data->cmacs);
    return 0;
  }

  data->oacs = cacs;
  return 0;
}


/* ---------------------------------------------------------------------------------
	section 12.2.2 Case 2: Only one side supports ACS change
   --------------------------------------------------------------------------------- */

int calc_OACS_case2(Data* data)
{
  CodecSet fixACS, scs, cscs;

  if ((data->lOM == 1) && (data->dOM == 0))
  {
    fixACS = data->dacs;
    scs    = data->lscs;
  }
  else if ((data->lOM == 0) && (data->dOM == 1))
  {
    fixACS = data->lacs;
    scs    = data->dscs;
  }
  else
  {
    fprintf(stderr, "ERROR in calc_OACS_case2()\n");
    exit(0);
  }

  cscs = fixACS & scs;
  
  if (!contiguous_subset(cscs, fixACS))
  {
    data->oacs = contiguous_subset_selection (cscs, data->lOM, data->dOM,
					                          data->lacs, data->dacs, data->cmacs);
    return 0;
  }

  if (nModes(cscs) > data->cmacs)
  {
    data->oacs = contiguous_subset_selection (cscs, data->lOM, data->dOM,
					                          data->lacs, data->dacs, data->cmacs);
    return 0;
  }

  data->oacs = cscs;

  return 0;
}


/* ---------------------------------------------------------------------------------
	section 12.4 Rules for contiguous subset selection                
   --------------------------------------------------------------------------------- */

CodecSet contiguous_subset_selection (CodecSet cs, int lOM, int dOM,
									  CodecSet lacs, CodecSet dacs, int cmacs)
{
  int mode;
  for (mode=mode_122; mode; mode>>=1)
  {
    if ( (lOM || contiguous_subset(cs, lacs)) &&
	     (dOM || contiguous_subset(cs, dacs)) && 
	     (nModes(cs) <= cmacs)
	   )
      break;

    cs &= mode ^ 0xFF;
  }
  return cs;
}



/* ---------------------------------------------------------------------------------
                                  terminal blocks
   ---------------------------------------------------------------------------------

int codec_mismatch_resolution(Data* data)
{
  if ( (data->lOM == 1) || (data->dOM == 1) ) 
	  printf("First Codec Mismatch Resolution, then TFO is possible.\n");

  else 
	  printf("TFO not possible, because no side can adapt.\n");

  return 0;
}
*/

/* ---------------------------------------------------------------------------------
                                  help routines
   --------------------------------------------------------------------------------- */

/* return a string with the name of an NB-mode */
char* modeName(CodecMode m)
{
  switch(m)
  {
	case mode_122: return "12.2";
	case mode_102: return "10.2";
	case mode_795: return "7.95";
	case mode_740: return "7.40";
	case mode_670: return "6.70";
	case mode_590: return "5.90";
	case mode_515: return "5.15";
	case mode_475: return "4.75";
	default:
	  fprintf(stderr, "ERROR in Codec Mode ()\n");	// only for debug purposes
	  return "????";
  }
}

/* return a string with the name of a WB-mode */
char* WBmodeName(CodecMode m)
{
  switch(m)
  {
	case mode_740: return "23.85";
	case mode_670: return "15.85";
	case mode_590: return "12.65";
	case mode_515: return " 8.85";
	case mode_475: return " 6.60";
	default:
	  fprintf(stderr, "ERROR in WB Codec Mode ()\n");	// only for debug purposes
	  return "????";
  }
}
	  

/* return a string with the name of a codec type */
char* codecName(CodecType t)
{
  switch(t)
  {											
	case   FR_AMR:     return "FR_AMR     ";
	case   HR_AMR:     return "HR_AMR     ";
	case  OHR_AMR:	   return "OHR_AMR    ";
	case UMTS_AMR:     return "UMTS_AMR   ";
	case UMTS_AMR_2:   return "UMTS_AMR_2 ";

	case  GSM_FR:      return "GSM_FR     ";
	case  GSM_HR:      return "GSM_HR     ";
	case  GSM_EFR:     return "GSM_EFR    ";

	case   FR_AMR_WB:  return "FR_AMR-WB  ";
	case  OHR_AMR_WB:  return "OHR_AMR-WB ";
	case  OFR_AMR_WB:  return "OFR_AMR-WB ";
	case UMTS_AMR_WB:  return "UMTS_AMR-WB";
	default:
	  fprintf(stderr, "ERROR in name of Codec Type()\n");	// only for debug purposes
	  return "????????????";
  }
}
