/*
******************************************************************************
*
*      File          : tfo_main.c We consider version handling no longer;
*      NOTE          : it is not expected that this SW will ever change,
*                      therefore all version handling is removed.
*      Version       : C.0.0, 02. Dec 2014, Karl Hellwig, Ericsson
*      Version       : B.0.0, 18. September 2012, Paolo Usai, Karl Hellwig
*      Version       : A.0.0, 13. April 2011, Karl Hellwig
*      Version       : 9.0.0, 11. December 2009, Paolo Usai
*      Version       : 8.1.0, 27. March 2009, Karl Hellwig	 <= Keep Track in helptext* 
*                      functional identical to 6.3.0
*      Version       : 7.0.0, 14. June 2007, Paolo Usai,
*      Version       : 6.4.0, 04. June 2007, Karl Hellwig, 
*      Version       : 6.3.0, 02. October 2006, Karl Hellwig
*      Version       : 6.2.0, 30. August 2005, Karl Hellwig
*      Version       : 2.2.0, 27. November 2000, Jochen Junkawitsch
*      Description   : C-Reference for TFO Decision Algorithm of 3GPP TS 28.062
*
******************************************************************************
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "tfo_decision.h"	/* KHe: consider version here no longer */
/* #include <iostream.h> */

int	helptext ()
{
	/*                           Update new Release Numbers not amymore                         */
	printf("=====================================================================\n");
	printf("TFO Decision Algorithm; C-Reference for TS 28.062, C.0.0 and beyond  \n");
	printf("                                                                     \n");
	printf("for GSM_FR     GSM_HR      GSM_EFR                                   \n");
	printf("    FR_AMR     HR_AMR      OHR_AMR     UMTS_AMR_2    UMTS_AMR        \n");
	printf("    FR_AMR-WB  OFR_AMR-WB  OHR_AMR-WB  UMTS_AMR-WB                   \n");
	printf("                                                                     \n");
	printf("This TFO Decision Algorithm determines, if TFO can be established    \n");
	printf("                                        and on which Configuration.  \n");
	printf("                                                                     \n");
	printf("Several input formats are supported (l_=local and d_=distant):       \n");
	printf("                                                                     \n");
	printf("l_Type      Preferred_Config.        d_Type     Preferred_Config.    \n");
	printf("l_Type      Preferred_Config.        d_Type     ACS      SCS  MACS OM\n");
	printf("l_Type      ACS      SCS   MACS OM   d_Type     ACS      SCS  MACS OM\n");
	printf("                                                                     \n");
	printf("Preferred_Config.: 0..15 for AMR and 0..5 for AMR-WB.                \n");
	printf("                                                                     \n");
	printf("ACS and SCS: highest ... lowest mode; x: mode present, -: not present\n");
	printf("MACS: 1..8; OM: n or y.                                              \n");
	printf("                                                                     \n");
	printf("Please separate parameters by blanks or tabs. Valid examples are:    \n");
	printf("                                                                     \n");
	printf("GSM_EFR     1                        GSM_EFR    1                    \n");
	printf("FR_AMR      12                       HR_AMR     10                   \n");
	printf("FR_AMR      x--x-x-x xxxxxxxx 4 y    HR_AMR     ---x-x-x --xxxxxx 3 n\n");
	printf("UMTS_AMR    x------- xxxxxxxx 1 n    UMTS_AMR_2 x------- xxxxxxxx 1 n\n");
	printf("UMTS_AMR_2  1                        HR_AMR     ---x-x-x --xxxxxx 3 n\n");
	printf("UMTS_AMR-WB 5                        FR_AMR-WB  0                    \n");
	printf("=====================================================================\n");
	return 1;
}

CodecSet s2cs(char* s)
/* ----------------------------------------------------------------------------------------------
   convert a ACS/SCS-string to the internal integer CodecSet (0..255) (example: "x--x-x-x" ->149)
   ---------------------------------------------------------------------------------------------- */
{
  int k;
  CodecSet m, cs = 0;
  
  if (strlen(s) != 8) return 0;
  
  for (k=0, m=mode_122; k<8; k++, m >>= 1)
    if ( (s[k] == 'X') || (s[k] == 'x') )
      cs |= m;
  
  return cs;
}
/* ----------------------------------------------------------------------------------------------
  Read Input Data from stdin 

  This routine reads the user input and validates it in a simple form.

  It can also process input parameter files with

  command_prompt> TFO <Input.txt >Output.txt

  In Feb 2005 this routine was extented to contain also

  - the short coding of TFO_Message as described in TS 28.062 table 7.11.3.1.3-2 
  - the short coding of TFO Message as described in TS 26.103 for AMR-WB
  - handle also OHR_AMR
  - handle the UMTS_AMR (in-)compatibility in all multi-mode Configurations.

  The help text is not repeated for every prompt, but only on demand: "h"elp

  The exit is explicitely coded with "e"xit 
  ---------------------------------------------------------------------------------------------- */

int getInputData (Data* data, int *FileProc)
{
  char buffer[255];
  char lscs[255], dscs[255];
  char lacs[255], dacs[255];
  int  lmacs, dmacs;
  char lType[255], dType[255];
  char lOM, dOM;


Input_Loop:

    data->cacs = data->iacs = data->oacs = data->cscs = data->lmacs = 0;

	printf(" \nYour input, please; (e=exit h=help ;=comment):\n\n");

Comment_Loop:

	if (!fgets(buffer, 255, stdin)) return 0;	// get input line. Exit, when the Input.Txt file is finished

/*	----------------------------------------------------------------------------------
	Check first for ";", because input files must be mirrored to the output 
	---------------------------------------------------------------------------------- */

	sscanf(buffer, "%s", lType);

    if (!strcmp(lType, ";")) *FileProc = 1;

	if (*FileProc) fputs(buffer, stdout);		// mirror the input file to the output file
    if (!strcmp(lType, ";")) goto Comment_Loop;	// comment lines need no processing

/*	----------------------------------------------------------------------------------
	Now check for 'e'xit or 'h'elp
	---------------------------------------------------------------------------------- */

    if ((sscanf(buffer, "%s", lType) == 1) && (!strcmp(lType, "e"))) 
	{
		printf(" OK exit now\n");
		return 0;
	}

    if ((sscanf(buffer, "%s", lType) == 1) && (!strcmp(lType, "h")))
	{
		helptext();
		goto Input_Loop;
	}

/************************************************************************************
*	Now check for the simplest, shortest input format:								*
*																					*
*	lType  Preferred_Configuration     dType  Preferred_Configuration				*
*																					*
*	The Preferred Configuration for AMR is mapped to the ACS, SCS, MACS and OM		*
*	as specified in the text below table 7.11.3.1.3-2.								*
*	The result is, however, that the TFO Decision sometimes determines				*
*	a Configuration that is NOT contained in this set of Preferred Configurations	*
*	=> discrepancy in the specification!											*
*																					*
*	The Preferred Configurations for AMR-WB are mapped to ACS, SCS, MACS and OM		*
*	as in case of a 5-mode Codec.													*
*	The text form of the TFO Decision for AMR-WB may need clarifications.			*
*																					*
************************************************************************************/

    if (sscanf(buffer, "%s    %d      %s     %d",
			           lType, &lmacs, dType, &dmacs) == 4)
	 {

	/********************************************************************************
	*	first check the local Codec Type											*
	*********************************************************************************/
 	if      (!strcmp(lType, "FR_AMR"     )) data->lType = FR_AMR;
	else if (!strcmp(lType, "HR_AMR"     )) data->lType = HR_AMR;
	else if (!strcmp(lType, "OHR_AMR"    )) data->lType = OHR_AMR;
	else if (!strcmp(lType, "UMTS_AMR"   )) data->lType = UMTS_AMR;
	else if (!strcmp(lType, "UMTS_AMR_2" )) data->lType = UMTS_AMR_2;
	else if (!strcmp(lType, "UMTS_AMR2"  )) data->lType = UMTS_AMR_2;

 	else if (!strcmp(lType, "FR_AMR-WB"  )) data->lType = FR_AMR_WB;
	else if (!strcmp(lType, "OFR_AMR-WB" )) data->lType = OFR_AMR_WB;
	else if (!strcmp(lType, "OHR_AMR-WB" )) data->lType = OHR_AMR_WB;
	else if (!strcmp(lType, "UMTS_AMR-WB")) data->lType = UMTS_AMR_WB;

 	else if (!strcmp(lType, "FR_AMR_WB"  )) data->lType = FR_AMR_WB;
	else if (!strcmp(lType, "OFR_AMR_WB" )) data->lType = OFR_AMR_WB;
	else if (!strcmp(lType, "OHR_AMR_WB" )) data->lType = OHR_AMR_WB;
	else if (!strcmp(lType, "UMTS_AMR_WB")) data->lType = UMTS_AMR_WB;

	else if (!strcmp(lType, "GSM_EFR"    )) data->lType = GSM_EFR;
	else if (!strcmp(lType, "GSM_FR"     )) data->lType = GSM_FR;
	else if (!strcmp(lType, "GSM_HR"     )) data->lType = GSM_HR;

	else    {fprintf(stderr,"\a\aERROR: unknown local Codec Type\n");	goto Input_Loop;}
  
	if ((data->lType == HR_AMR) && ((lmacs == 6) || (lmacs == 7)||(lmacs > 10)) || (lmacs > 15))
	{ fprintf(stderr,"\a\aERROR in local Codec Type: This Preferred Configuration is not allowed \n");
		 goto Input_Loop;
	}

	if (data->lType == FR_AMR     ||       
        data->lType == HR_AMR     ||
        data->lType == OHR_AMR    ||
        data->lType == UMTS_AMR   || 
        data->lType == UMTS_AMR_2   )
	{
      if (lmacs == 0)  {data->lacs =      data->lscs = 1;   data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 1)  {data->lacs =      data->lscs = 149; data->lmacs = 4; data->lOM = 0;}
      if (lmacs == 2)  {data->lacs =      data->lscs = 4;   data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 3)  {data->lacs =      data->lscs = 8;   data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 4)  {data->lacs =      data->lscs = 16;  data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 5)  {data->lacs =      data->lscs = 32;  data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 6)  {data->lacs =      data->lscs = 64;  data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 7)  {data->lacs =      data->lscs = 128; data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 8)  {data->lacs =      data->lscs = 5;   data->lmacs = 2; data->lOM = 0;}
      if (lmacs == 9)  {data->lacs =      data->lscs = 13;  data->lmacs = 3; data->lOM = 0;}
      if (lmacs == 10) {data->lacs =      data->lscs = 29;  data->lmacs = 4; data->lOM = 0;}
      if (lmacs == 11) {data->lacs = 29;  data->lscs = 253; data->lmacs = 4; data->lOM = 1;}
      if (lmacs == 12) {data->lacs =      data->lscs = 77;  data->lmacs = 4; data->lOM = 0;}
      if (lmacs == 13) {data->lacs = 77;  data->lscs = 253; data->lmacs = 4; data->lOM = 1;}
      if (lmacs == 14) {data->lacs =      data->lscs = 165; data->lmacs = 4; data->lOM = 0;}
      if (lmacs == 15) {data->lacs = 165; data->lscs = 253; data->lmacs = 4; data->lOM = 1;}

	  if ((data->lType == HR_AMR) && (data-> lscs == 149)) // correct Preferred Configuration 1 for HR_AMR
	  {
					    data->lacs =      data->lscs = 21;  data->lmacs = 3; data->lOM = 0;
	  }
	}

	if (data->lType == FR_AMR_WB  ||       
        data->lType == OHR_AMR_WB   )
	{
		if (lmacs > 1) 
		{	
			fprintf(stderr,"\a\aERROR in local Codec Type: This Preferred Configuration is not allowed \n");
			goto Input_Loop;
		}
		if (lmacs == 0)  {data->lacs =      data->lscs =  7; data->lmacs = 3; data->lOM = 0;}
		if (lmacs == 1)  {data->lacs =   7; data->lscs =  7; data->lmacs = 3; data->lOM = 0;}
	}

    if (data->lType == OFR_AMR_WB ||
        data->lType == UMTS_AMR_WB  ) 
	{
		if (lmacs > 5) 
		{
			fprintf(stderr,"\a\aERROR in local Codec Type: This Preferred Configuration is not allowed \n");
			goto Input_Loop;
		}
		if (lmacs == 0)  {data->lacs =      data->lscs =  7; data->lmacs = 3; data->lOM = 0;}
		if (lmacs == 1)  {data->lacs =   7; data->lscs = 31; data->lmacs = 4; data->lOM = 1;}
		if (lmacs == 2)  {data->lacs =      data->lscs = 15; data->lmacs = 4; data->lOM = 0;}
		if (lmacs == 3)  {data->lacs =  15; data->lscs = 31; data->lmacs = 4; data->lOM = 1;}
		if (lmacs == 4)  {data->lacs =      data->lscs = 23; data->lmacs = 4; data->lOM = 0;}
		if (lmacs == 5)  {data->lacs =  23; data->lscs = 31; data->lmacs = 4; data->lOM = 1;}
	}
	else if (data->lType == GSM_FR ||
			 data->lType == GSM_HR ||
			 data->lType == GSM_EFR  ) 
			{ data->lOM = data->lscs = data->lacs = 0; data->lmacs = 1; }

	/********************************************************************************
	*	now check the distant Codec Type											*
	*********************************************************************************/

	if      (!strcmp(dType, "FR_AMR"     )) data->dType = FR_AMR;
	else if (!strcmp(dType, "HR_AMR"     )) data->dType = HR_AMR;
	else if (!strcmp(dType, "OHR_AMR"    )) data->dType = OHR_AMR;
	else if (!strcmp(dType, "UMTS_AMR"   )) data->dType = UMTS_AMR;
	else if (!strcmp(dType, "UMTS_AMR_2" )) data->dType = UMTS_AMR_2;
	else if (!strcmp(dType, "UMTS_AMR2"  )) data->dType = UMTS_AMR_2;
	
 	else if (!strcmp(dType, "FR_AMR-WB"  )) data->dType = FR_AMR_WB;
	else if (!strcmp(dType, "OFR_AMR-WB" )) data->dType = OFR_AMR_WB;
	else if (!strcmp(dType, "OHR_AMR-WB" )) data->dType = OHR_AMR_WB;
	else if (!strcmp(dType, "UMTS_AMR-WB")) data->dType = UMTS_AMR_WB;
	
 	else if (!strcmp(dType, "FR_AMR_WB"  )) data->dType = FR_AMR_WB;
	else if (!strcmp(dType, "OFR_AMR_WB" )) data->dType = OFR_AMR_WB;
	else if (!strcmp(dType, "OHR_AMR_WB" )) data->dType = OHR_AMR_WB;
	else if (!strcmp(dType, "UMTS_AMR_WB")) data->dType = UMTS_AMR_WB;

	else if (!strcmp(dType, "GSM_EFR"    )) data->dType = GSM_EFR;
	else if (!strcmp(dType, "GSM_FR"     )) data->dType = GSM_FR;
	else if (!strcmp(dType, "GSM_HR"     )) data->dType = GSM_HR;

	else    {fprintf(stderr,"\a\aERROR: unknown distant codec type\n");	goto Input_Loop;}

	if ( (data->dType == HR_AMR) && ((dmacs == 6) || (dmacs == 7)||(dmacs > 10)) || (dmacs > 15) )
	{
		fprintf(stderr,"\a\aERROR in distant Codec Type: This Preferred Configuration is not allowed \n");
		goto Input_Loop;
	}

	if (data->dType == FR_AMR     ||       
        data->dType == HR_AMR     ||
        data->dType == OHR_AMR    ||
        data->dType == UMTS_AMR   || 
        data->dType == UMTS_AMR_2   )
	{
      if (dmacs == 0)  {data->dacs =      data->dscs = 1;    data->dmacs = 1; data->dOM = 0;}
      if (dmacs == 1)  {data->dacs =      data->dscs = 149;  data->dmacs = 4; data->dOM = 0;}
      if (dmacs == 2)  {data->dacs =      data->dscs = 4;    data->dmacs = 1; data->dOM = 0;}
      if (dmacs == 3)  {data->dacs =      data->dscs = 8;    data->dmacs = 1; data->dOM = 0;}
      if (dmacs == 4)  {data->dacs =      data->dscs = 16;   data->dmacs = 1; data->dOM = 0;}
      if (dmacs == 5)  {data->dacs =      data->dscs = 32;   data->dmacs = 1; data->dOM = 0;}
      if (dmacs == 6)  {data->dacs =      data->dscs = 64;   data->dmacs = 1; data->dOM = 0;}
      if (dmacs == 7)  {data->dacs =      data->dscs = 128;  data->dmacs = 1; data->dOM = 0;}
      if (dmacs == 8)  {data->dacs =      data->dscs = 5;    data->dmacs = 2; data->dOM = 0;}
      if (dmacs == 9)  {data->dacs =      data->dscs = 13;   data->dmacs = 3; data->dOM = 0;}
      if (dmacs == 10) {data->dacs =      data->dscs = 29;   data->dmacs = 4; data->dOM = 0;}
      if (dmacs == 11) {data->dacs = 29;  data->dscs = 253;  data->dmacs = 4; data->dOM = 1;}
      if (dmacs == 12) {data->dacs =      data->dscs = 77;   data->dmacs = 4; data->dOM = 0;}
      if (dmacs == 13) {data->dacs = 77;  data->dscs = 253;  data->dmacs = 4; data->dOM = 1;}
      if (dmacs == 14) {data->dacs =      data->dscs = 165;  data->dmacs = 4; data->dOM = 0;}
      if (dmacs == 15) {data->dacs = 165; data->dscs = 253;  data->dmacs = 4; data->dOM = 1;}

	  if ((data->dType == HR_AMR) && (data-> dscs == 149)) // correct Preferred Configuration 1 for HR_AMR
	  {
					    data->dacs =      data->dscs = 21;  data->dmacs = 3; data->dOM = 0;
	  }

	}
	if (data->dType ==  FR_AMR_WB ||       
        data->dType == OHR_AMR_WB   )
	{
		if (dmacs > 1) 
		{
			fprintf(stderr,"\a\aERROR in distant Codec Type: This Preferred Configuration is not allowed \n");
			goto Input_Loop;
		}
		if (dmacs == 0)  {data->dacs =      data->dscs =  7; data->dmacs = 3; data->dOM = 0;}
		if (dmacs == 1)  {data->dacs =   7; data->dscs =  7; data->dmacs = 3; data->dOM = 0;}
	}

    if (data->dType ==  OFR_AMR_WB ||
        data->dType == UMTS_AMR_WB   ) 
	{
		if (dmacs > 5) 
		{
			fprintf(stderr,"\a\aERROR in distant Codec Type: This Preferred Configuration is not allowed \n");
			goto Input_Loop;
		}
		if (dmacs == 0)  {data->dacs =      data->dscs =  7; data->dmacs = 3; data->dOM = 0;}
		if (dmacs == 1)  {data->dacs =   7; data->dscs = 31; data->dmacs = 4; data->dOM = 1;}
		if (dmacs == 2)  {data->dacs =      data->dscs = 15; data->dmacs = 4; data->dOM = 0;}
		if (dmacs == 3)  {data->dacs =  15; data->dscs = 31; data->dmacs = 4; data->dOM = 1;}
		if (dmacs == 4)  {data->dacs =      data->dscs = 23; data->dmacs = 4; data->dOM = 0;}
		if (dmacs == 5)  {data->dacs =  23; data->dscs = 31; data->dmacs = 4; data->dOM = 1;}
	}
	else if (data->dType == GSM_FR ||
			 data->dType == GSM_HR ||
			 data->dType == GSM_EFR  ) 
			{ data->dOM = data->dscs = data->dacs = 0; data->dmacs = 1; }

	return 1;	// this is the return from input when the shortest form is validated
	}

/************************************************************************************
*	now check for the combination of short and longest input format:				*
*	lType lpreferred 	dType dACS dSCS dMACS dOM									*
************************************************************************************/

    if (sscanf(buffer, "%s    %d      %s     %s    %s    %d      %c",
			           lType, &lmacs, dType, dacs, dscs, &dmacs, &dOM) == 7)
	{

	/********************************************************************************
	*	first check the local Codec Type											*
	*********************************************************************************/
 	if      (!strcmp(lType, "FR_AMR"     )) data->lType = FR_AMR;
	else if (!strcmp(lType, "HR_AMR"     )) data->lType = HR_AMR;
	else if (!strcmp(lType, "OHR_AMR"    )) data->lType = OHR_AMR;
	else if (!strcmp(lType, "UMTS_AMR"   )) data->lType = UMTS_AMR;
	else if (!strcmp(lType, "UMTS_AMR_2" )) data->lType = UMTS_AMR_2;
	else if (!strcmp(lType, "UMTS_AMR2"  )) data->lType = UMTS_AMR_2;

 	else if (!strcmp(lType, "FR_AMR-WB"  )) data->lType = FR_AMR_WB;
	else if (!strcmp(lType, "OFR_AMR-WB" )) data->lType = OFR_AMR_WB;
	else if (!strcmp(lType, "OHR_AMR-WB" )) data->lType = OHR_AMR_WB;
	else if (!strcmp(lType, "UMTS_AMR-WB")) data->lType = UMTS_AMR_WB;

 	else if (!strcmp(dType, "FR_AMR_WB"  )) data->dType = FR_AMR_WB;
	else if (!strcmp(dType, "OFR_AMR_WB" )) data->dType = OFR_AMR_WB;
	else if (!strcmp(dType, "OHR_AMR_WB" )) data->dType = OHR_AMR_WB;
	else if (!strcmp(dType, "UMTS_AMR_WB")) data->dType = UMTS_AMR_WB;

	else if (!strcmp(lType, "GSM_EFR"    )) data->lType = GSM_EFR;
	else if (!strcmp(lType, "GSM_FR"     )) data->lType = GSM_FR;
	else if (!strcmp(lType, "GSM_HR"     )) data->lType = GSM_HR;

	else    {fprintf(stderr,"\a\aERROR: unknown local Codec Type\n");	goto Input_Loop;}
  
	if ( (data->lType == HR_AMR) && ((lmacs == 6) || (lmacs == 7)||(lmacs > 10)) || (lmacs > 15) )
	{
		fprintf(stderr,"\a\aERROR in local Codec Type: This Preferred Configuration is not allowed \n");
		goto Input_Loop;
	}

	if (data->lType == FR_AMR     ||       
        data->lType == HR_AMR     ||
        data->lType == OHR_AMR    ||
        data->lType == UMTS_AMR   || 
        data->lType == UMTS_AMR_2   )
	{
      if (lmacs == 0)  {data->lacs =      data->lscs = 1;   data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 1)  {data->lacs =      data->lscs = 149; data->lmacs = 4; data->lOM = 0;}
      if (lmacs == 2)  {data->lacs =      data->lscs = 4;   data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 3)  {data->lacs =      data->lscs = 8;   data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 4)  {data->lacs =      data->lscs = 16;  data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 5)  {data->lacs =      data->lscs = 32;  data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 6)  {data->lacs =      data->lscs = 64;  data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 7)  {data->lacs =      data->lscs = 128; data->lmacs = 1; data->lOM = 0;}
      if (lmacs == 8)  {data->lacs =      data->lscs = 5;   data->lmacs = 2; data->lOM = 0;}
      if (lmacs == 9)  {data->lacs =      data->lscs = 13;  data->lmacs = 3; data->lOM = 0;}
      if (lmacs == 10) {data->lacs =      data->lscs = 29;  data->lmacs = 4; data->lOM = 0;}
      if (lmacs == 11) {data->lacs = 29;  data->lscs = 253; data->lmacs = 4; data->lOM = 1;}
      if (lmacs == 12) {data->lacs =      data->lscs = 77;  data->lmacs = 4; data->lOM = 0;}
      if (lmacs == 13) {data->lacs = 77;  data->lscs = 253; data->lmacs = 4; data->lOM = 1;}
      if (lmacs == 14) {data->lacs =      data->lscs = 165; data->lmacs = 4; data->lOM = 0;}
      if (lmacs == 15) {data->lacs = 165; data->lscs = 253; data->lmacs = 4; data->lOM = 1;}

	  if ((data->lType == HR_AMR) && (data-> lscs == 149)) // correct Preferred Configuration 1 for HR_AMR
	  {
					    data->lacs =      data->lscs = 21;  data->lmacs = 3; data->lOM = 0;
	  }

	}
	if (data->lType == FR_AMR_WB  ||       
        data->lType == OHR_AMR_WB   )
	{
		if (lmacs > 1) 
		{
			fprintf(stderr,"\a\aERROR in local Codec Type: This Preferred Configuration is not allowed \n");
			goto Input_Loop;
		}
		if (lmacs == 0)  {data->lacs =      data->lscs =  7; data->lmacs = 3; data->lOM = 0;}
		if (lmacs == 1)  {data->lacs =   7; data->lscs =  7; data->lmacs = 3; data->lOM = 0;}
	}

    if (data->lType == OFR_AMR_WB ||
        data->lType == UMTS_AMR_WB  ) 
	{
		if (lmacs > 5) 
		{
			fprintf(stderr,"\a\aERROR in local Codec Type: This Preferred Configuration is not allowed \n");
			goto Input_Loop;
		}
		if (lmacs == 0)  {data->lacs =      data->lscs =  7; data->lmacs = 3; data->lOM = 0;}
		if (lmacs == 1)  {data->lacs =   7; data->lscs = 31; data->lmacs = 4; data->lOM = 1;}
		if (lmacs == 2)  {data->lacs =      data->lscs = 15; data->lmacs = 4; data->lOM = 0;}
		if (lmacs == 3)  {data->lacs =  15; data->lscs = 31; data->lmacs = 4; data->lOM = 1;}
		if (lmacs == 4)  {data->lacs =      data->lscs = 23; data->lmacs = 4; data->lOM = 0;}
		if (lmacs == 5)  {data->lacs =  23; data->lscs = 31; data->lmacs = 4; data->lOM = 1;}
	}
	else if (data->lType == GSM_FR ||
			 data->lType == GSM_HR ||
			 data->lType == GSM_EFR  ) 
			{ data->lOM = data->lscs = data->lacs = 0; data->lmacs = 1; }

	/********************************************************************************
	*	now check the distant Codec Type											*
	*********************************************************************************/
	goto long_distant;
	}

/************************************************************************************
*	Now check for the longest input format:											*
*																					*
*	lType lACS lSCS lMACS lOM	dType dACS dSCS dMACS dOM							*
*																					*
*	This is basically as it was until Version 4.5.0									*
*	Just the sequence of parameters was adapted to the sequence in TS 26.103		*
*	The OHR_AMR is now included.													*
*																					*
*	Please note: this input format is NOT specified for AMR-WB!						*
************************************************************************************/

	if (sscanf(buffer, "%s    %s     %s   %d      %c    %s     %s    %s    %d      %c",
			         lType, lacs, lscs, &lmacs, &lOM, dType, dacs, dscs, &dmacs, &dOM) == 10)
	{

	/********************************************************************************
	*	now check the local Codec Type												*
	*********************************************************************************/

	if      (!strcmp(lType, "FR_AMR"    )) data->lType = FR_AMR;
	else if (!strcmp(lType, "HR_AMR"    )) data->lType = HR_AMR;
	else if (!strcmp(lType, "OHR_AMR"   )) data->lType = OHR_AMR;
	else if (!strcmp(lType, "UMTS_AMR"  )) data->lType = UMTS_AMR;
	else if (!strcmp(lType, "UMTS_AMR_2")) data->lType = UMTS_AMR_2;
	else if (!strcmp(lType, "UMTS_AMR2" )) data->lType = UMTS_AMR_2;

	else if (!strcmp(lType, "GSM_EFR"   )) data->lType = GSM_EFR;
	else if (!strcmp(lType, "GSM_FR"    )) data->lType = GSM_FR;
	else if (!strcmp(lType, "GSM_HR"    )) data->lType = GSM_HR;
	else    {fprintf(stderr,"\a\aERROR: local Codec Type not allowed here\n");				goto Input_Loop;}
  
	if (data->lType == FR_AMR     ||       
        data->lType == HR_AMR     ||
        data->lType == OHR_AMR    ||
        data->lType == UMTS_AMR   || 
        data->lType == UMTS_AMR_2   )
	{

      if (!(data->lacs = s2cs(lacs))) {fprintf(stderr,"\a\aERROR: l_ACS not 8 chars long\n"); goto Input_Loop;}
	  if (!(data->lscs = s2cs(lscs))) {fprintf(stderr,"\a\aERROR: l_SCS not 8 chars long\n"); goto Input_Loop;}

	  if ( (data->lType == FR_AMR  ||       
            data->lType == HR_AMR  ||
            data->lType == OHR_AMR   ) && (lmacs > 4)
		 ){fprintf(stderr,"\a\aERROR: l_MACS bigger than 4\n");								goto Input_Loop;}

      if   (lmacs > 0 && lmacs <= 8) data->lmacs = lmacs;
      else {fprintf(stderr,"\a\aERROR: l_MACS not between 1 and 8\n");						goto Input_Loop;}

      if      (lOM == 'y') data->lOM = 1;
      else if (lOM == 'n') data->lOM = 0;
      else    {fprintf(stderr,"\a\aERROR: l_OM not 'y' or 'n'\n");							goto Input_Loop;}

	  if ((data->lacs & data->lscs) != data->lacs)
	  {fprintf(stderr,"\a\aERROR: l_ACS not in l_SCS\n");									goto Input_Loop;}

      if (data->lmacs < nModes(data->lacs))
	  {fprintf(stderr,"\a\aERROR: More modes in l_ACS than what l_MACS allows \n");			goto Input_Loop;}

      if ( (data->lType == HR_AMR) && ( (data->lscs > 63) || (data->lacs > 63) ) )
	  {fprintf(stderr,"\a\aERROR: Highest two AMR modes not allowed in local HR_AMR:\n");	goto Input_Loop;}

	}

	/********************************************************************************
	*	now check the distant Codec Type											*
	*	this part is also used in the combination >short form> - <long form>		*
	*********************************************************************************/
long_distant:

	if      (!strcmp(dType, "FR_AMR"    )) data->dType = FR_AMR;
	else if (!strcmp(dType, "HR_AMR"    )) data->dType = HR_AMR;
	else if (!strcmp(dType, "OHR_AMR"   )) data->dType = OHR_AMR;
	else if (!strcmp(dType, "UMTS_AMR"  )) data->dType = UMTS_AMR;
	else if (!strcmp(dType, "UMTS_AMR_2")) data->dType = UMTS_AMR_2;
	else if (!strcmp(dType, "UMTS_AMR2" )) data->dType = UMTS_AMR_2;
	else if (!strcmp(dType, "GSM_EFR"   )) data->dType = GSM_EFR;
	else if (!strcmp(dType, "GSM_FR"    )) data->dType = GSM_FR;
	else if (!strcmp(dType, "GSM_HR"    )) data->dType = GSM_HR;
	else    {fprintf(stderr,"\a\aERROR: distant Codec Type not allowed here\n");			goto Input_Loop;}

	if (data->dType == FR_AMR     ||       
        data->dType == HR_AMR     ||
        data->dType == OHR_AMR    ||
        data->dType == UMTS_AMR   || 
        data->dType == UMTS_AMR_2   )
	{
      if (!(data->dacs = s2cs(dacs))) {fprintf(stderr,"\a\aERROR: d_ACS not 8 chars long\n"); goto Input_Loop;}
	  if (!(data->dscs = s2cs(dscs))) {fprintf(stderr,"\a\aERROR: d_SCS not 8 chars long\n"); goto Input_Loop;}

	  if ( (data->dType == FR_AMR  ||       
            data->dType == HR_AMR  ||
            data->dType == OHR_AMR   ) && (dmacs > 4)
		 ){fprintf(stderr,"\a\aERROR: d_MACS bigger than 4\n");								goto Input_Loop;}

     if   (dmacs > 0 && dmacs <= 8) data->dmacs = dmacs;
     else {fprintf(stderr,"\a\aERROR: d_MACS not between 1 and 8\n");						goto Input_Loop;}

     if      (dOM == 'y') data->dOM = 1;
     else if (dOM == 'n') data->dOM = 0;
     else    {fprintf(stderr,"\a\aERROR: d_OM not 'y' or 'n'\n");							goto Input_Loop;}

	 if ((data->dacs & data->dscs) != data->dacs)
	 {fprintf(stderr,"\a\aERROR: d_ACS not in d_SCS\n");									goto Input_Loop;}

     if (data->dmacs < nModes(data->dacs))
     {fprintf(stderr,"\a\aERROR: More modes in d_ACS than what d_MACS allows \n");			goto Input_Loop;}

     if ( (data->dType == HR_AMR) && ( (data->dscs > 63) || (data->dacs > 63) ) )
	 {fprintf(stderr,"\a\aERROR: Highest two AMR modes not allowed in distant HR_AMR:\n");	goto Input_Loop;}

	}

  // reset unnecessary parameters for simple Codec Types

  if ( data->lType == GSM_FR    || 
	   data->lType == GSM_HR    ||
	   data->lType == GSM_EFR     
	 ){data->lOM = data->lscs = data->lacs = 0; data->lmacs = 1;}

  if ( data->dType == GSM_FR    || 
	   data->dType == GSM_HR    ||
	   data->dType == GSM_EFR     
	 ){data->dOM = data->dscs = data->dacs = 0; data->dmacs = 1;}

  return 1;	// this is the return when the long form and the combination is validated.
 }

 fprintf(stderr,"\a\aERROR! This combination of parameters is not supported by this program\n");
 goto Input_Loop;

}
/************************************************************************************/
/* End of getInputData																*/
/************************************************************************************/



/************************************************************************************/
/* Print Data to stdout																*/
/* Prints all details for AMR and AMR-WB											*/
/* not needed for simple codec types such as GSM_FR, GSM_HR, GSM_EFR				*/
/************************************************************************************/
int printData(Data* data)
{
  CodecSet m;
  
  /* Two compatible AMR codec types? */
  if (
	  data->lType == UMTS_AMR   && data->dType == UMTS_AMR   ||
	  
	  data->lType == UMTS_AMR_2 && data->dType == UMTS_AMR_2 ||
      data->lType == UMTS_AMR_2 && data->dType == FR_AMR     ||
      data->lType == UMTS_AMR_2 && data->dType == OHR_AMR    ||
      data->lType == UMTS_AMR_2 && data->dType == HR_AMR     ||

      data->lType == FR_AMR     && data->dType == UMTS_AMR_2 ||
      data->lType == FR_AMR     && data->dType == FR_AMR     ||
      data->lType == FR_AMR     && data->dType == OHR_AMR    ||
      data->lType == FR_AMR     && data->dType == HR_AMR     ||

      data->lType == OHR_AMR    && data->dType == UMTS_AMR_2 ||
      data->lType == OHR_AMR    && data->dType == FR_AMR     ||
      data->lType == OHR_AMR    && data->dType == OHR_AMR    ||
      data->lType == OHR_AMR    && data->dType == HR_AMR     ||

      data->lType == HR_AMR     && data->dType == UMTS_AMR_2 ||
      data->lType == HR_AMR     && data->dType == FR_AMR     ||
      data->lType == HR_AMR     && data->dType == OHR_AMR    ||
      data->lType == HR_AMR     && data->dType == HR_AMR     ||

      data->lType == UMTS_AMR_2 && data->dType == UMTS_AMR   ||
      data->lType == UMTS_AMR   && data->dType == UMTS_AMR_2   )
  {
	  printf("\n\t%s\t\t\t\t%s\n", codecName(data->lType), codecName(data->dType));
	  printf("\tMACS = %d\t\t\t\tMACS = %d\n", data->lmacs, data->dmacs);
      printf("\tOM   = %s\t\t\t\tOM   = %s\n", data->lOM ? "yes" : "no", data->dOM ? "yes" : "no");
	  printf("\n\tSCS\tACS\tiACS\toACS\tcSCS\tACS\tSCS\n");

  	  for (m=mode_122; m; m>>=1)
	  {
		printf("%s\t%c\t%c\t%c\t%c\t%c\t%c\t%c\n",
				 modeName((CodecMode)m),
				 data->lscs & m ? 'x' : '-',
				 data->lacs & m ? 'x' : '-',
				 data->cacs & m ? 'x' : '-',
				 data->oacs & m ? 'x' : '-',
				 data->cscs & m ? 'x' : '-',
				 data->dacs & m ? 'x' : '-',
				 data->dscs & m ? 'x' : '-');
	  }
  }

  /* Two compatible AMR-WB codec types? */
  if (data->lType == UMTS_AMR_WB && data->dType == UMTS_AMR_WB ||
	  data->lType == UMTS_AMR_WB && data->dType ==  OFR_AMR_WB ||
      data->lType == UMTS_AMR_WB && data->dType ==  OHR_AMR_WB ||
      data->lType == UMTS_AMR_WB && data->dType ==   FR_AMR_WB ||

	  data->lType ==  OFR_AMR_WB && data->dType == UMTS_AMR_WB ||
	  data->lType ==  OFR_AMR_WB && data->dType ==  OFR_AMR_WB ||
      data->lType ==  OFR_AMR_WB && data->dType ==  OHR_AMR_WB ||
      data->lType ==  OFR_AMR_WB && data->dType ==   FR_AMR_WB ||
 
	  data->lType ==  OHR_AMR_WB && data->dType == UMTS_AMR_WB ||
	  data->lType ==  OHR_AMR_WB && data->dType ==  OFR_AMR_WB ||
      data->lType ==  OHR_AMR_WB && data->dType ==  OHR_AMR_WB ||
      data->lType ==  OHR_AMR_WB && data->dType ==   FR_AMR_WB ||

	  data->lType ==   FR_AMR_WB && data->dType == UMTS_AMR_WB ||
	  data->lType ==   FR_AMR_WB && data->dType ==  OFR_AMR_WB ||
      data->lType ==   FR_AMR_WB && data->dType ==  OHR_AMR_WB ||
      data->lType ==   FR_AMR_WB && data->dType ==   FR_AMR_WB   )
  {
	  printf("\n\t%s\t\t\t\t%s\n", codecName(data->lType), codecName(data->dType));
	  printf("\tMACS = %d\t\t\t\tMACS = %d\n", data->lmacs, data->dmacs);
      printf("\tOM   = %s\t\t\t\tOM   = %s\n", data->lOM ? "yes" : "no", data->dOM ? "yes" : "no");
	  printf("\n\tSCS\tACS\tiACS\toACS\tcSCS\tACS\tSCS\n");
  
	  for (m=mode_740; m; m>>=1)	// the 5 lower AMR modenames are "misused" also for WB modes
	  {
		printf("%s\t%c\t%c\t%c\t%c\t%c\t%c\t%c\n",
				 WBmodeName((CodecMode)m),
				 data->lscs & m ? 'x' : '-',
				 data->lacs & m ? 'x' : '-',
				 data->cacs & m ? 'x' : '-',
				 data->oacs & m ? 'x' : '-',
				 data->cscs & m ? 'x' : '-',
				 data->dacs & m ? 'x' : '-',
				 data->dscs & m ? 'x' : '-');
	  }
  }

  return 0;
}
/************************************************************************************/
/* End of Print Data to stdout														*/
/************************************************************************************/


/************************************************************************************/
/* Begin of Main																	*/
/************************************************************************************/
int main()
{
  Data			a;
  Data* data = &a;
  int  FileProc = 0;				// Marker for ";" entries: these indicate an input file

  	helptext();

    while (getInputData(data, &FileProc) != 0)
	{
	  printf("--------------------------------------------------------------------- \n");
//      printf("Result of the TFO Decision: \n");
      printf("==> ");

	  tfo_decision_main(data);	// determines and prints the main TFO decision for all Codec Types and Codec Families
	  printData        (data);	// prints details for all AMR and AMR-WB Codec Types
	}

  printf("===========================================================================\n");
  printf("Exiting TFO C-Reference                                                    \n");
  printf("===========================================================================\n");
  return 0;
}
/************************************************************************************/
/* End of Main																		*/
/************************************************************************************/
